#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Contains helper functions for the class `SPXtlm`."""

from __future__ import annotations

__all__ = ["UNITS_DICT", "convert_hk"]

from enum import Enum

import numpy as np
from numpy import ma

# This dictionary is only valid when raw counts are converted to physical units
UNITS_DICT = {
    "ADC1_GAIN": "Volt",
    "ADC1_OFFSET": "Volt",
    "ADC1_REF": "Volt",
    "ADC1_T": "K",
    "ADC1_VCC": "Volt",
    "ADC2_GAIN": "Volt",
    "ADC2_OFFSET": "Volt",
    "ADC2_REF": "Volt",
    "ADC2_T": "K",
    "ADC2_VCC": "Volt",
    "DEM_I": "mA",
    "DEM_V": "Volt",
    "DET_T": "K",
    "HTR1_DUTYCYCL": "%",
    "HTR1_I": "mA",
    "HTR2_DUTYCYCL": "%",
    "HTR2_I": "mA",
    "HTR3_DUTYCYCL": "%",
    "HTR3_I": "mA",
    "HTR4_DUTYCYCL": "%",
    "HTR4_I": "mA",
    "HTRGRP1_V": "Volt",
    "HTRGRP2_V": "Volt",
    "ICU_1P2V_I": "mA",
    "ICU_1P2V_V": "Volt",
    "ICU_3P3V_I": "mA",
    "ICU_3P3V_V": "Volt",
    "ICU_4P0V_I": "mA",
    "ICU_4P0V_V": "Volt",
    "ICU_4V_T": "K",
    "ICU_5P0V_I": "mA",
    "ICU_5P0V_V": "Volt",
    "ICU_5V_T": "K",
    "ICU_DIGV_T": "K",
    "ICU_HG1_T": "K",
    "ICU_HG2_T": "K",
    "ICU_MCU_T": "K",
    "ICU_MID_T": "K",
    "LED1_ANODE_V": "Volt",
    "LED1_CATH_V": "Volt",
    "LED1_I": "mA",
    "LED2_ANODE_V": "Volt",
    "LED2_CATH_V": "Volt",
    "LED2_I": "mA",
    "TS1_DEM_N_T": "K",
    "TS2_HOUSING_N_T": "K",
    "TS3_RADIATOR_N_T": "K",
    "TS4_DEM_R_T": "K",
    "TS5_HOUSING_R_T": "K",
    "TS6_RADIATOR_R_T": "K",
}


# - helper functions ------------------------
def exp_spex_det_t(raw_data: np.ndarray) -> np.ndarray:
    """Convert Detector Temperature Sensor to [K]."""
    res = np.empty(raw_data.size, dtype=float)
    mask = raw_data < 400
    res[mask] = 1.224 * raw_data[mask] - 17.05
    res[~mask] = 0.6426 * raw_data[~mask] - 145.57
    res[res > 325] = np.nan
    return res


def exp_spex_thermistor(raw_data: np.ndarray) -> np.ndarray:
    """Convert readouts of the Temperature Sensors to [K].

    Notes
    -----
    - TS1 DEM Nominal temperature
    - TS2 Housing Nominal Temperature
    - TS3 Radiator Nominal Temperature
    - TS4 DEM Redundant Temperature
    - TS5 Housing Redundant Temperature
    - TS6 Radiator Redundant Temperature*

    """
    coefficients = (294.34, 272589.0, 1.5173e-15, 5.73666e-19, 5.11328e-20)
    buff = ma.masked_array(raw_data / 256, mask=raw_data == 0)
    buff = (
        coefficients[0]
        + coefficients[1] / buff
        - coefficients[2] * buff**4
        + (coefficients[3] - coefficients[4] * ma.log(buff)) * buff**5
    )
    buff[raw_data == 0] = np.nan
    return buff.data


def poly_spex_icuhk_internaltemp(raw_data: np.ndarray) -> np.ndarray:
    """Convert readouts of temperature sensors on ICU power supplies to [K].

    Notes
    -----
    - ICU V5 supply temperature
    - ICU V4 supply temperature
    - ICU HtrG1 supply temperature
    - ICU HtrG2 supply temperature
    - ICU MidBoard temperature
    - ICU MCU-RAM temperature
    - ICU 1V2, 3V3 supply temperature

    """
    coefficients = (273.15, 0.0625)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_icuhk_internaltemp2(raw_data: np.ndarray) -> np.ndarray:
    """Convert readings of ICU bus voltages to Voltages.

    Notes
    -----
    - ICU 4.0 Volt bus voltage
    - ICU 3.3 Volt bus voltage
    - ICU 1.2 Volt bus voltage
    - DEM power supply

    """
    coefficients = (0, 0.001)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_htr_v(raw_data: np.ndarray) -> np.ndarray:
    """Convert Heater Bus voltages to Volt."""
    coefficients = (0, 0.01 / 3)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_htr1_p(raw_data: np.ndarray) -> np.ndarray:
    """Convert Heater1 Current to Watt."""
    resistance = 238.3
    return resistance * (raw_data / 1000) ** 2


def poly_spex_htr2_p(raw_data: np.ndarray) -> np.ndarray:
    """Convert Heater2 Current to Watt."""
    resistance = 212.5
    return resistance * (raw_data / 1000) ** 2


def poly_spex_htr3_p(raw_data: np.ndarray) -> np.ndarray:
    """Convert Heater3 Current to Watt."""
    resistance = 237.7
    return resistance * (raw_data / 1000) ** 2


def poly_spex_htr4_p(raw_data: np.ndarray) -> np.ndarray:
    """Convert Heater4 Current to Watt."""
    resistance = 213.2
    return resistance * (raw_data / 1000) ** 2


def poly_spex_dutycycle(raw_data: np.ndarray) -> np.ndarray:
    """Convert Heater Controller Duty Cycle output to percent."""
    coefficients = (0, 0.1)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_led_anode_v(raw_data: np.ndarray) -> np.ndarray:
    """Convert LED Anode voltage to Volt."""
    coefficients = (0, 0.000000623703003)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_led_cath_v(raw_data: np.ndarray) -> np.ndarray:
    """Convert LED Cathode voltage to Volt."""
    coefficients = (0, 0.000000415802001953)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_led_i(raw_data: np.ndarray) -> np.ndarray:
    """Convert LED current to milli-Ampere."""
    coefficients = (0, 0.0000030307446495961334745762711864407)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_adc_vcc(raw_data: np.ndarray) -> np.ndarray:
    """Convert ADC Analog VCC reading to Volt."""
    coefficients = (0, 0.00000127156575520833333)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_adc_gain(raw_data: np.ndarray) -> np.ndarray:
    """Convert ADC Analog VCC reading to Volt."""
    coefficients = (0, 0.000000127156575520833333)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_adc_t(raw_data: np.ndarray) -> np.ndarray:
    """Convert ADC1 Temperature reading to [K]."""
    coefficients = (-0.25, 0.0007385466842651367)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_adc_offset(raw_data: np.ndarray) -> np.ndarray:
    """Convert ADC offset (?) to Voltage."""
    coefficients = (0, 0.000415802001953)
    return coefficients[0] + coefficients[1] * raw_data


def poly_spex_dem_i(raw_data: np.ndarray) -> np.ndarray:
    """Convert DEM Supply current to milli-Ampere."""
    coefficients = (0, 0.2417)
    return coefficients[0] + coefficients[1] * raw_data


class Enable(Enum):
    """Possible values of the DEM SpaceWire link status."""

    INACTIVE = 0
    ACTIVE = 1


class EnaDis(Enum):
    """Possible values of the LED commanded enable state."""

    OFF = 0
    ON = 1


class IcuMode(Enum):
    """Possible values of the ICU mode."""

    IDLE = 0
    OPER = 1


class IsEna(Enum):
    """Possible values of the Power Sense."""

    DISABLED = 0
    ENABLED = 1


class OcdReg(Enum):
    """Possible values of the OCD state."""

    PWR_OK = 0
    OC2V1 = 1
    OC3V3 = 2
    OC2V1_AND_3V3 = 3


class TimeMsg(Enum):
    """Received status of external PPS or Time message."""

    MISSING = 0
    RECEIVED = 1


class WriteProt(Enum):
    """Status of Flash Write protection (both banks)."""

    BANK0_DISABLED = 0
    BANK0_ENABLED = 1
    BANK1_DISABLED = 16
    BANK1_ENABLED = 17


# - exported functions ----------------------
def convert_hk(key: str, raw_data: np.ndarray) -> np.ndarray:
    """Convert a DemHK or NomHK parameter to physical units."""
    conv_dict = {
        "DET_T": exp_spex_det_t,
        "TS1_DEM_N_T": exp_spex_thermistor,
        "TS2_HOUSING_N_T": exp_spex_thermistor,
        "TS3_RADIATOR_N_T": exp_spex_thermistor,
        "TS4_DEM_R_T": exp_spex_thermistor,
        "TS5_HOUSING_R_T": exp_spex_thermistor,
        "TS6_RADIATOR_R_T": exp_spex_thermistor,
        "ADC1_GAIN": poly_spex_adc_gain,
        "ADC2_GAIN": poly_spex_adc_gain,
        "ADC1_OFFSET": poly_spex_adc_offset,
        "ADC2_OFFSET": poly_spex_adc_offset,
        "ADC1_T": poly_spex_adc_t,
        "ADC2_T": poly_spex_adc_t,
        "ADC1_REF": poly_spex_adc_vcc,
        "ADC1_VCC": poly_spex_adc_vcc,
        "ADC2_REF": poly_spex_adc_vcc,
        "ADC2_VCC": poly_spex_adc_vcc,
        "DEM_I": poly_spex_dem_i,
        "HTR1_DUTYCYCL": poly_spex_dutycycle,
        "HTR2_DUTYCYCL": poly_spex_dutycycle,
        "HTR3_DUTYCYCL": poly_spex_dutycycle,
        "HTR4_DUTYCYCL": poly_spex_dutycycle,
        "HTR1_POWER": poly_spex_htr1_p,
        "HTR2_POWER": poly_spex_htr2_p,
        "HTR3_POWER": poly_spex_htr3_p,
        "HTR4_POWER": poly_spex_htr4_p,
        "HTRGRP1_V": poly_spex_htr_v,
        "HTRGRP2_V": poly_spex_htr_v,
        "ICU_4V_T": poly_spex_icuhk_internaltemp,
        "ICU_5V_T": poly_spex_icuhk_internaltemp,
        "ICU_DIGV_T": poly_spex_icuhk_internaltemp,
        "ICU_HG1_T": poly_spex_icuhk_internaltemp,
        "ICU_HG2_T": poly_spex_icuhk_internaltemp,
        "ICU_MCU_T": poly_spex_icuhk_internaltemp,
        "ICU_MID_T": poly_spex_icuhk_internaltemp,
        "DEM_V": poly_spex_icuhk_internaltemp2,
        "ICU_1P2V_V": poly_spex_icuhk_internaltemp2,
        "ICU_3P3V_V": poly_spex_icuhk_internaltemp2,
        "ICU_4P0V_V": poly_spex_icuhk_internaltemp2,
        "ICU_5P0V_V": poly_spex_icuhk_internaltemp2,
        "LED1_ANODE_V": poly_spex_led_anode_v,
        "LED2_ANODE_V": poly_spex_led_anode_v,
        "LED1_CATH_V": poly_spex_led_cath_v,
        "LED2_CATH_V": poly_spex_led_cath_v,
        "LED1_I": poly_spex_led_i,
        "LED2_I": poly_spex_led_i,
    }

    func = conv_dict.get(key)
    if func is not None:
        return func(raw_data)

    return raw_data



#!/usr/bin/env python3
#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Python script to generate a report from the CKD in a CKD product."""

from __future__ import annotations

import argparse

from moniplot.lib.fig_info import FIGinfo
from moniplot.mon_plot import MONplot
from pyspex.ckd_io import CKDio

# - global parameters ------------------------------


# - local functions --------------------------------
def init_plot_file(key: str, ckd: CKDio, ckd_ref: CKDio | None = None) -> tuple:
    """Initialize CKD report."""
    fig_info_in = FIGinfo()
    fig_info_in.add(
        "processor_version", (ckd.processor_version, ckd.git_commit), fmt="{} ({})"
    )
    fig_info_in.add("processing_date", ckd.date_created())

    # open CKD product to be used as reference
    if ckd_ref is not None:
        fig_info_in.add(
            "reference_version",
            (ckd_ref.processor_version, ckd_ref.git_commit),
            fmt="{} ({})",
        )

    # open CKD report
    ckd_date = ckd.date_created(compact=True)
    plot = MONplot(f"spx1_ckd_report_{key}_{ckd_date}.pdf")
    plot.set_institute("SRON")

    return plot, fig_info_in


def add_dark_figs(ckd: CKDio, ckd_ref: CKDio | None = None) -> None:
    """Generate figures of Dark CKD."""
    dark_ckd = ckd.dark()
    if dark_ckd is None:
        return

    ckd_version = "v1" if "dark_offset" in dark_ckd.data_vars else "v2"

    plot, fig_info_in = init_plot_file("dark", ckd, ckd_ref)
    plot.set_caption("SPEXone Dark CKD")
    if ckd_version == "v1":
        plot.draw_signal(
            dark_ckd["dark_offset"], title="dark offset", fig_info=fig_info_in.copy()
        )
        plot.draw_hist(
            dark_ckd["dark_offset"],
            bins=161,
            vrange=[-5.5, 2.5],
            title="dark offset",
            fig_info=fig_info_in.copy(),
        )
    else:
        plot.draw_signal(
            dark_ckd["offset_long"], title="offset (long)", fig_info=fig_info_in.copy()
        )
        plot.draw_hist(
            dark_ckd["offset_long"],
            bins=161,
            vrange=[-5.5, 2.5],
            title="offset (long)",
            fig_info=fig_info_in.copy(),
        )
        plot.draw_signal(
            dark_ckd["offset_short"],
            title="offset (short)",
            fig_info=fig_info_in.copy(),
        )
        plot.draw_hist(
            dark_ckd["offset_short"],
            bins=161,
            vrange=[-5.5, 2.5],
            title="offset (short)",
            fig_info=fig_info_in.copy(),
        )
    plot.draw_signal(
        dark_ckd["dark_current"], title="dark current", fig_info=fig_info_in.copy()
    )
    plot.draw_hist(
        dark_ckd["dark_current"],
        bins=101,
        vrange=[1.5, 6.5],
        title="dark current",
        fig_info=fig_info_in.copy(),
    )

    ref_ckd = ckd_ref.dark() if ckd_ref is not None else None
    if ref_ckd is None:
        plot.close()
        return

    ref_version = "v1" if "dark_offset" in ref_ckd.data_vars else "v2"

    if ckd_version == "v1":
        if ref_version == "v1":
            plot.draw_signal(
                dark_ckd["dark_offset"] - ref_ckd["dark_offset"],
                title="dark offset - reference",
                fig_info=fig_info_in.copy(),
                zscale="diff",
            )
        else:
            plot.draw_signal(
                dark_ckd["dark_offset"] - ref_ckd["offset_long"],
                title="dark offset - reference",
                fig_info=fig_info_in.copy(),
                zscale="diff",
            )
    else:
        if ref_version == "v1":
            plot.draw_signal(
                dark_ckd["offset_long"] - ref_ckd["dark_offset"],
                title="offset (long) - reference",
                fig_info=fig_info_in.copy(),
                zscale="diff",
            )
        else:
            plot.draw_signal(
                dark_ckd["offset_short"] - ref_ckd["offset_short"],
                title="offset (short) - reference",
                fig_info=fig_info_in.copy(),
                zscale="diff",
            )
            plot.draw_signal(
                dark_ckd["offset_long"] - ref_ckd["offset_long"],
                title="offset (long) - reference",
                fig_info=fig_info_in.copy(),
                zscale="diff",
            )

    plot.draw_signal(
        dark_ckd["dark_current"] - ref_ckd["dark_current"],
        title="dark current - reference",
        fig_info=fig_info_in.copy(),
        zscale="diff",
    )
    plot.close()


def add_noise_figs(ckd: CKDio, ckd_ref: CKDio | None = None) -> None:
    """Generate figures of Noise CKD."""
    noise_ckd = ckd.noise()
    if noise_ckd is None:
        return

    plot, fig_info_in = init_plot_file("noise", ckd, ckd_ref)
    plot.set_caption("SPEXone Noise CKD")
    g_str = "inverse conversion gain"
    plot.draw_signal(
        1 / noise_ckd["g"], vrange=[11.5, 15], title=g_str, fig_info=fig_info_in.copy()
    )
    plot.draw_hist(
        1 / noise_ckd["g"],
        bins=161,
        vrange=[9.5, 17.5],
        title=g_str,
        fig_info=fig_info_in.copy(),
    )
    n_str = "read noise"
    plot.draw_signal(noise_ckd["n"], title=n_str, fig_info=fig_info_in.copy())
    plot.draw_hist(
        noise_ckd["n"],
        bins=161,
        vrange=[0, 8],
        title=n_str,
        fig_info=fig_info_in.copy(),
    )

    ref_ckd = ckd_ref.noise() if ckd_ref is not None else None
    if ref_ckd is not None:
        plot.draw_signal(
            noise_ckd["g"] - ref_ckd["g"],
            title=g_str + " - reference",
            fig_info=fig_info_in.copy(),
            zscale="diff",
        )
        plot.draw_signal(
            noise_ckd["n"] - ref_ckd["n"],
            title=n_str + " - reference",
            fig_info=fig_info_in.copy(),
            zscale="diff",
        )
    plot.close()


def add_nlin_figs(ckd: CKDio, ckd_ref: CKDio | None = None) -> None:
    """Generate figures of non-linearity CKD."""
    nlin_ckd = ckd.nlin()
    if nlin_ckd is None:
        return

    plot, fig_info_in = init_plot_file("nlin", ckd, ckd_ref)
    plot.set_caption("SPEXone non-linearity CKD")
    # if 'A' in nlin_ckd:
    #    plot.draw_signal(nlin_ckd['A'], title='parameter A',
    #                     fig_info=fig_info_in.copy())
    #    plot.draw_signal(nlin_ckd['B'], title='parameter B',
    #                     fig_info=fig_info_in.copy())
    #    plot.draw_signal(nlin_ckd['C'], title='parameter C',
    #                     fig_info=fig_info_in.copy())
    if "c" in nlin_ckd:
        plot.draw_signal(
            nlin_ckd["f1"], title="parameter f1", fig_info=fig_info_in.copy()
        )
        plot.draw_signal(
            nlin_ckd["f2"], title="parameter f2", fig_info=fig_info_in.copy()
        )
        plot.draw_signal(
            nlin_ckd["c"], title="sigmoid parameter c", fig_info=fig_info_in.copy()
        )
        plot.draw_signal(
            nlin_ckd["r0"], title="sigmoid parameter r0", fig_info=fig_info_in.copy()
        )
        plot.draw_signal(
            nlin_ckd["r1"], title="sigmoid parameter r1", fig_info=fig_info_in.copy()
        )
        plot.draw_signal(
            nlin_ckd["r2"], title="sigmoid parameter r2", fig_info=fig_info_in.copy()
        )
        plot.draw_signal(
            nlin_ckd["r3"], title="sigmoid parameter r3", fig_info=fig_info_in.copy()
        )
        plot.draw_signal(
            nlin_ckd["r4"], title="sigmoid parameter r4", fig_info=fig_info_in.copy()
        )
        plot.draw_signal(
            nlin_ckd["m0"], title="sigmoid parameter m0", fig_info=fig_info_in.copy()
        )
        # plot.draw_signal(nlin_ckd['m1'], title='sigmoid parameter m1',
        #                 fig_info=fig_info_in.copy())
        # plot.draw_signal(nlin_ckd['m2'], title='sigmoid parameter m2',
        #                 fig_info=fig_info_in.copy())

    ref_ckd = ckd_ref.nlin() if ckd_ref is not None else None
    if ref_ckd is not None:
        # plot commands...
        pass
    plot.close()


def add_prnu_figs(ckd: CKDio, ckd_ref: CKDio | None = None) -> None:
    """Generate figures of PRNU CKD."""
    prnu_ckd = ckd.prnu()
    if prnu_ckd is None:
        return

    ref_ckd = ckd_ref.prnu() if ckd_ref is not None else None

    plot, fig_info_in = init_plot_file("prnu", ckd, ckd_ref)
    plot.set_caption("SPEXone PRNU CKD")

    prnu_str = "Pixel Response Non-Uniformity"
    plot.draw_signal(prnu_ckd, fig_info=fig_info_in.copy(), title=prnu_str)
    if ref_ckd is not None:
        prnu_str = "Pixel Response Non-Uniformity (reference)"
        plot.draw_signal(ref_ckd, fig_info=fig_info_in.copy(), title=prnu_str)
    plot.draw_hist(
        prnu_ckd,
        bins=201,
        vrange=[0.95, 1.05],
        title=prnu_str,
        fig_info=fig_info_in.copy(),
    )
    if ref_ckd is not None:
        prnu_str = "Pixel Response Non-Uniformity (reference)"
        plot.draw_hist(
            ref_ckd,
            bins=201,
            vrange=[0.95, 1.05],
            title=prnu_str,
            fig_info=fig_info_in.copy(),
        )
        prnu_str = "Pixel Response Non-Uniformity"
        plot.draw_signal(
            prnu_ckd - ref_ckd,
            title=prnu_str + " - reference",
            fig_info=fig_info_in.copy(),
            zscale="diff",
        )
    plot.close()


# def add_fov_figs(ckd: CKDio, ckd_ref: CKDio | None = None) -> None:
#    """Generate figures of Field-of-View CKD.
#    """
#    fov_ckd = ckd.fov()
#    if fov_ckd is None:
#        return
#
#    plot, fig_info_in = init_plot_file('fov', ckd, ckd_ref)
#    plot.set_caption('SPEXone field-of-view CKD')
#    # plot commands...
#
#    ref_ckd = ckd_ref.fov() if ckd_ref is not None else None
#    if ref_ckd is not None:
#        # plot commands...
#        pass
#    plot.close()


# def add_swath_figs(ckd: CKDio, ckd_ref: CKDio | None = None) -> None:
#    """Generate figures of Swath CKD.
#    """
#    swath_ckd = ckd.swath()
#    if swath_ckd is None:
#        return
#
#    plot, fig_info_in = init_plot_file('swath', ckd, ckd_ref)
#    plot.set_caption('SPEXone Swath CKD')
#    # plot commands...
#
#    ref_ckd = ckd_ref.swath() if ckd_ref is not None else None
#    if ref_ckd is not None:
#        # plot commands...
#        pass
#    plot.close()


def add_wave_figs(ckd: CKDio, ckd_ref: CKDio | None = None) -> None:
    """Generate figures of Wavelength CKD."""
    wave_ckd = ckd.wavelength()
    if wave_ckd is None:
        return

    plot, fig_info_in = init_plot_file("wave", ckd, ckd_ref)
    plot.set_caption("SPEXone Wavelength CKD")
    wave_s_str = "wavelength grid of the S spectra"
    plot.draw_signal(
        wave_ckd["wave_full"][0, ...], fig_info=fig_info_in.copy(), title=wave_s_str
    )
    wave_p_str = "wavelength grid of the P spectra"
    plot.draw_signal(
        wave_ckd["wave_full"][1, ...], fig_info=fig_info_in.copy(), title=wave_p_str
    )
    wave_str = "common wavelength grid for S and P"
    plot.draw_signal(
        wave_ckd["wave_common"], fig_info=fig_info_in.copy(), title=wave_str
    )

    ref_ckd = ckd_ref.wavelength() if ckd_ref is not None else None
    if ref_ckd is not None:
        plot.draw_signal(
            wave_ckd["wave_common"] - ref_ckd["wave_common"],
            title=wave_str + " - reference",
            fig_info=fig_info_in.copy(),
            zscale="diff",
        )
        plot.draw_signal(
            wave_ckd["wave_full"][0, ...] - ref_ckd["wave_full"][0, ...],
            title=wave_s_str + " - reference",
            fig_info=fig_info_in.copy(),
            zscale="diff",
        )
        plot.draw_signal(
            wave_ckd["wave_full"][1, ...] - ref_ckd["wave_full"][1, ...],
            title=wave_p_str + " - reference",
            fig_info=fig_info_in.copy(),
            zscale="diff",
        )
    plot.close()


def add_rad_figs(ckd: CKDio, ckd_ref: CKDio | None = None) -> None:
    """Generate figures of Radiometric CKD."""
    rad_ckd = ckd.radiometric()
    if rad_ckd is None:
        return

    plot, fig_info_in = init_plot_file("rad", ckd, ckd_ref)
    plot.set_caption("SPEXone Radiometric CKD")
    rad_s_str = rad_ckd.attrs["long_name"] + " (S)"
    plot.draw_signal(rad_ckd[:, 0, :], fig_info=fig_info_in.copy(), title=rad_s_str)
    rad_p_str = rad_ckd.attrs["long_name"] + " (P)"
    plot.draw_signal(rad_ckd[:, 1, :], fig_info=fig_info_in.copy(), title=rad_p_str)

    ref_ckd = ckd_ref.radiometric() if ckd_ref is not None else None
    if ref_ckd is not None:
        plot.draw_signal(
            rad_ckd[:, 0, :] - ref_ckd[:, 0, :],
            title=rad_s_str + " - reference",
            fig_info=fig_info_in.copy(),
            zscale="diff",
        )
        plot.draw_signal(
            rad_ckd[:, 1, :] - ref_ckd[:, 1, :],
            title=rad_p_str + " - reference",
            fig_info=fig_info_in.copy(),
            zscale="diff",
        )
    plot.close()


def add_pol_figs(ckd: CKDio, ckd_ref: CKDio | None = None) -> None:
    """Generate figures of Polarimetric CKD."""
    pol_ckd = ckd.polarimetric()
    if pol_ckd is None:
        return

    plot, fig_info_in = init_plot_file("pol", ckd, ckd_ref)
    plot.set_caption("SPEXone Polarimetric CKD")
    pol_q_str = pol_ckd["pol_m_q"].attrs["long_name"]
    plot.draw_signal(pol_ckd["pol_m_q"], fig_info=fig_info_in.copy(), title=pol_q_str)
    pol_u_str = pol_ckd["pol_m_u"].attrs["long_name"]
    plot.draw_signal(pol_ckd["pol_m_u"], fig_info=fig_info_in.copy(), title=pol_u_str)
    pol_t_str = pol_ckd["pol_m_t"].attrs["long_name"]
    plot.draw_signal(pol_ckd["pol_m_t"], fig_info=fig_info_in.copy(), title=pol_t_str)

    ref_ckd = ckd_ref.polarimetric() if ckd_ref is not None else None
    if ref_ckd is not None:
        plot.draw_signal(
            pol_ckd["pol_m_q"] - ref_ckd["pol_m_q"],
            title=pol_q_str + " - reference",
            fig_info=fig_info_in.copy(),
            zscale="diff",
        )
        plot.draw_signal(
            pol_ckd["pol_m_u"] - ref_ckd["pol_m_u"],
            title=pol_u_str + " - reference",
            fig_info=fig_info_in.copy(),
            zscale="diff",
        )
        plot.draw_signal(
            pol_ckd["pol_m_t"] - ref_ckd["pol_m_t"],
            title=pol_t_str + " - reference",
            fig_info=fig_info_in.copy(),
            zscale="diff",
        )
    plot.close()


# - main function ----------------------------------
def main() -> None:
    """Generate a report on SPEXone CKD."""
    # parse command-line parameters
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description="Generate a PDF report from CKD in a SPEXone CKD product",
    )

    parser.add_argument(
        "--ref_ckd_file", default=None, help="name of reference CKD product"
    )
    parser.add_argument("ckd_file", help="name of CKD product")
    args = parser.parse_args()

    # open CKD product
    ckd = CKDio(args.ckd_file)

    # open CKD product to be used as reference
    ckd_ref = None
    if args.ref_ckd_file is not None:
        ckd_ref = CKDio(args.ref_ckd_file)

    # add CKD's to report
    add_dark_figs(ckd, ckd_ref)
    add_noise_figs(ckd, ckd_ref)
    add_nlin_figs(ckd, ckd_ref)
    add_prnu_figs(ckd, ckd_ref)
    # add_fov_figs(ckd, ckd_ref)
    # add_swath_figs(ckd, ckd_ref)
    add_wave_figs(ckd, ckd_ref)
    add_rad_figs(ckd, ckd_ref)
    add_pol_figs(ckd, ckd_ref)


# --------------------------------------------------
if __name__ == "__main__":
    main()



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Contains the class `TIFio` to read simulated SPEXone measurements."""

from __future__ import annotations

__all__ = ["TIFio"]

from pathlib import Path

import numpy as np

try:
    import pytiff
except ModuleNotFoundError:
    FOUND_PYTIFF = False
else:
    FOUND_PYTIFF = True

# - global parameters ------------------------------


# - local functions --------------------------------


# - class TIFio -------------------------
if FOUND_PYTIFF:

    class TIFio:
        """Class to read SPEXone instrument simulator output.

        Parameters
        ----------
        hdr_file : str
        inp_tif : bool
        lineskip : bool

        Examples
        --------
        Read a SPEXone TIFF file::

        >>  tif = TIFio(filename)
        >>  print(tif.header())
        >>  print(tif.tags()[0])
        >>  print(tif.images().shape)

        """

        def __init__(
            self: TIFio, hdr_file: str, inp_tif: bool = False, lineskip: bool = False
        ) -> None:
            """Initialize TIFio object."""
            if not Path(hdr_file).is_file():
                raise FileNotFoundError(f"file {hdr_file} not found")

            # initialize class-attributes
            self.filename = Path(hdr_file)
            self.dir_name = Path(hdr_file).parent
            self.__stem = Path(hdr_file).stem
            self.__header = None
            self.inp_tif = inp_tif
            self.lineskip = lineskip

        # --------------------------------------------------
        def header(self: TIFio) -> dict:
            """Read header as dictionary."""
            res = {}
            with self.filename.open(encoding="ascii", errors="ignore") as fp:
                res["history"] = fp.readline()[:-1]
                _ = fp.readline()
                for line in fp:
                    if line == "-\n":
                        break
                    key, value = line[:-1].split(":")
                    if value.strip(" ") != "None":
                        res[key.rstrip(" ")] = value.strip(" ")

                # read Spectral data stimulus
                line = fp.readline()
                buff = line[:-1].split("[")
                key = buff[0].rstrip(" ")
                if key == "Spectral data stimulus":
                    ds_sets = buff[1].replace("]", "").split(", ")

                    ds_dict = {}
                    for name in ds_sets:
                        line = fp.readline()
                        if not line:
                            break
                        ds_dict[name] = np.array(line[:-1].split(","), dtype=float)

                    if ds_dict:
                        res[key] = ds_dict

            self.__header = res
            return res

        # --------------------------------------------------
        def tags(self: TIFio) -> list:
            """Return TIFF tags as dictionary."""
            if self.__header is None:
                self.header()
            n_frame = int(self.__header["Number of measurements"])

            res = []
            for num in range(n_frame):
                tif_path = self.dir_name / f"{self.__stem}_{num}.tif"
                with pytiff.Tiff(str(tif_path)) as handle:
                    res.append(handle.read_tags())

            return res

        # --------------------------------------------------
        def images(self: TIFio, n_frame: int | None = None) -> np.ndarray:
            """Return TIFF data as numpy array.

            Parameters
            ----------
            n_frame :  int, optional
               Distribute coadded signal (32-bit) over n_frame images (16-bit)

            """
            if self.__header is None:
                self.header()

            if self.inp_tif:
                tif_path = self.dir_name / f"{self.__stem}_inp.tif"
                with pytiff.Tiff(str(tif_path)) as handle:
                    data = handle[:]

                if n_frame is None:
                    n_frame = 1 + (data.max() // 0xFFFF)
                elif n_frame < 1 + (data.max() // 0xFFFF):
                    print("[WARNING]: n_frame too small - precision will be lost")

                if n_frame == 1:
                    return data.astype("u2")

                frames = np.zeros((n_frame,) + data.shape, dtype="u2")
                frames += data // n_frame
                diff = data - np.sum(frames, axis=0)
                for img in frames:
                    mask = diff > 0
                    img[mask] += 1
                    diff[mask] -= 1

                return frames

            # convert regular TIFF files
            n_frame = int(self.__header["Number of measurements"])

            if self.lineskip:
                tif_fmt = str(self.dir_name / "{}_lineskip_{}.tif")
            else:
                tif_fmt = str(self.dir_name / "{}_{}.tif")

            res = []
            for num in range(n_frame):
                with pytiff.Tiff(tif_fmt.format(self.__stem, num)) as handle:
                    res.append(handle[:])

            return np.array(res)



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Contains the class CKDio to read SPEXone CKD.

References
----------
* https://spexone-cal-doc.readthedocs.io/en/latest/


"""

from __future__ import annotations

__all__ = ["CKDio"]

from datetime import UTC, datetime
from typing import TYPE_CHECKING

import h5py
import xarray as xr
from moniplot.image_to_xarray import h5_to_xr

if TYPE_CHECKING:
    from pathlib import Path

# - global parameters ------------------------------

# - local functions --------------------------------


# - class CKDio -------------------------
class CKDio:
    """Defines a class to read SPEXone CKD parameters.

    Parameters
    ----------
    ckd_file :  str
        Name of CKD file
    verbose :  bool, default=False
        Be verbose

    Examples
    --------
    Read several CKD parameters:

    >>> with CKDio(Path('SPX1_CKD.nc')) as ckd:
    >>>    dark = ckd.dark()
    >>>    fov = ckd.fov()

    """

    def __init__(self: CKDio, ckd_file: Path, verbose: bool = False) -> None:
        """Initialize class attributes."""
        self.verbose = verbose

        # open access to CKD product
        self.fid = h5py.File(ckd_file, "r")
        if "processor_configuration" not in self.fid:
            raise RuntimeError("SPEXone CKD product corrupted?")

    def __enter__(self: CKDio) -> CKDio:
        """Initiate the context manager."""
        return self

    def __exit__(self: CKDio, *args: str) -> bool:
        """Exit the context manager."""
        self.close()
        return False  # any exception is raised by the with statement.

    def close(self: CKDio) -> None:
        """Make sure that we close all resources."""
        if self.fid is not None:
            self.fid.close()

    @property
    def processor_version(self: CKDio) -> str:
        """Return the version of the `spexone_cal` program."""
        # pylint: disable=no-member
        return self.fid.attrs["processor_version"].decode()

    def date_created(self: CKDio, compact: bool = False) -> str:
        """Return creation date of the CKD product.

        Parameters
        ----------
        compact :  bool
           return date in iso-format if not compact else return 'YYYYmmddHHMMSS'

        """
        # pylint: disable=no-member
        date_str = self.fid.attrs["date_created"].decode()
        date_t = datetime.strptime(date_str, "%Y %B %d %a %Z%z %H:%M:%S")
        if compact:
            return date_t.astimezone(tz=UTC).strftime("%Y%m%d%H%M%S")

        return date_t.astimezone(tz=UTC).isoformat()[:-6]

    @property
    def git_commit(self: CKDio) -> str:
        """Return git hash of repository `spexone_cal`, used to generate the CKD."""
        # pylint: disable=no-member
        return self.fid.attrs["git_commit"].decode()

    def dark(self: CKDio) -> xr.Dataset | None:
        """Read Dark CKD.

        Returns
        -------
        xarray.Dataset
           parameters of the SPEXone Dark CKD

        """
        try:
            gid = self.fid["DARK"]
        except KeyError:
            return None
        res = ()
        if "dark_offset" in gid:
            res += (h5_to_xr(gid["dark_offset"]),)
        else:
            res += (h5_to_xr(gid["offset_long"]),)
            res += (h5_to_xr(gid["offset_short"]),)
        res += (h5_to_xr(gid["dark_current"]),)
        return xr.merge(res, combine_attrs="drop_conflicts")

    def noise(self: CKDio) -> xr.Dataset | None:
        """Read Noise CKD.

        Returns
        -------
        xarray.Dataset
           parameters of the SPEXone Noise CKD

        """
        try:
            gid = self.fid["NOISE"]
        except KeyError:
            return None
        res = ()
        res += (h5_to_xr(gid["g"]),)
        res += (h5_to_xr(gid["n"]),)
        return xr.merge(res, combine_attrs="drop_conflicts")

    def nlin(self: CKDio) -> xr.Dataset | None:
        """Read non-linearity CKD.

        Returns
        -------
        xarray.Dataset
           parameters of the SPEXone non-linearity CKD

        """
        try:
            gid = self.fid["NON_LINEARITY"]
            sigmoidal = "A" in gid
        except KeyError:
            return None

        res = ()
        if sigmoidal:
            res += (h5_to_xr(gid["A"]),)
            res += (h5_to_xr(gid["B"]),)
            res += (h5_to_xr(gid["C"]),)
            if "/DEBUG/NON_LINEARITY" in self.fid:
                gid = self.fid["/DEBUG/NON_LINEARITY"]
                res += (h5_to_xr(gid["f1"]),)
                res += (h5_to_xr(gid["f2"]),)
                res += (h5_to_xr(gid["c"]),)
                res += (h5_to_xr(gid["r0"]),)
                res += (h5_to_xr(gid["r1"]),)
                res += (h5_to_xr(gid["r2"]),)
                res += (h5_to_xr(gid["r3"]),)
                res += (h5_to_xr(gid["r4"]),)
                res += (h5_to_xr(gid["m0"]),)
                # res += (h5_to_xr(gid['m1']),)
                # res += (h5_to_xr(gid['m2']),)
        else:
            res += (h5_to_xr(gid["nonlin_order"]),)
            res += (h5_to_xr(gid["nonlin_knots"]),)
            res += (h5_to_xr(gid["nonlin_exptimes"]),)
            res += (h5_to_xr(gid["nonlin_signal_scale"]),)
            res += (h5_to_xr(gid["nonlin_fit"]),)
        return xr.merge(res, combine_attrs="drop_conflicts")

    def prnu(self: CKDio) -> xr.DataArray | None:
        """Read PRNU CKD.

        Returns
        -------
        xr.DataArray
           parameters of the SPEXone PRNU CKD

        """
        try:
            gid = self.fid["PRNU"]
        except KeyError:
            return None
        return h5_to_xr(gid["prnu"])

    def fov(self: CKDio) -> xr.Dataset | None:
        """Read field-of-view CKD.

        Returns
        -------
        xarray.Dataset
           parameters of the SPEXone field-of-view CKD

        """
        try:
            gid = self.fid["FIELD_OF_VIEW"]
        except KeyError:
            return None
        res = ()
        res += (h5_to_xr(gid["fov_nfov_vp"]),)
        res += (h5_to_xr(gid["fov_ifov_start_vp"]),)
        res += (h5_to_xr(gid["fov_act_angles"]),)
        res += (h5_to_xr(gid["fov_ispat"]),)
        return xr.merge(res, combine_attrs="drop_conflicts")

    def wavelength(self: CKDio) -> xr.Dataset | None:
        """Read Wavelength CKD.

        Returns
        -------
        xarray.Dataset
           parameters of the SPEXone Wavelength CKD

        """
        try:
            gid = self.fid["WAVELENGTH"]
        except KeyError:
            return None
        res = ()
        # Before radiometric calibration S and P have separate wavelength grids
        res += (h5_to_xr(gid["wave_full"]),)
        # After radiometric calibration S and P are interpolated to a common
        # wavelength grid.
        res += (h5_to_xr(gid["wave_common"]),)
        return xr.merge(res, combine_attrs="drop_conflicts")

    def radiometric(self: CKDio) -> xr.DataArray | None:
        """Read Radiometric CKD.

        Returns
        -------
        xr.DataArray
           parameters of the SPEXone Radiometric CKD

        """
        try:
            gid = self.fid["RADIOMETRIC"]
        except KeyError:
            return None
        return h5_to_xr(gid["rad_spectra"])

    def polarimetric(self: CKDio) -> xr.Dataset | None:
        """Read polarimetric CKD.

        Returns
        -------
        xarray.Dataset
           parameters of the SPEXone polarimetric CKD

        """
        try:
            gid = self.fid["POLARIMETRIC"]
        except KeyError:
            return None
        res = ()
        res += (h5_to_xr(gid["pol_m_q"]),)
        res += (h5_to_xr(gid["pol_m_u"]),)
        res += (h5_to_xr(gid["pol_m_t"]),)
        return xr.merge(res, combine_attrs="drop_conflicts")



#!/usr/bin/env python3
#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Create or update SPEXone binning-table CKD.

- Input is binning table information in CSV format.
- The binning-table CKD is written in netCDF4 format.
"""

from __future__ import annotations

import argparse
from pathlib import Path

import numpy as np
from pyspex.binning_tables import BinningTables

# - global parameters ------------------------------


# - local functions --------------------------------


# - main function ----------------------------------
def main() -> None:
    """Generate SPEXone binning tables.

    Create or update SPEXone binning-table CKD.
    Input is binning table information in CSV format.
    The binning-table CKD is written in netCDF4 format.
    """
    parser = argparse.ArgumentParser(
        description="build a database with SPEXone binning tables"
    )
    parser.add_argument(
        "--verbose",
        default=False,
        action="store_true",
        help="be verbose, default be silent",
    )
    parser.add_argument(
        "--validity_start",
        default=None,
        type=str,
        help=("validity start of the binning tables" " (format: yyyymmddTHHMMSS)"),
    )
    parser.add_argument(
        "--table_id",
        type=str,
        default="1",
        help="start value table_id or comma separated list",
    )
    parser.add_argument(
        "--lineskip", default=None, help="provide lineskip arrays per binning table"
    )
    parser.add_argument(
        "file_list", nargs="+", help="provide binning table(s) in csv-format"
    )
    args = parser.parse_args()
    if args.verbose:
        print(args)

    # define a table_id per binning table
    table_id_list = args.table_id.split(",")
    if len(table_id_list) == 1:
        offs = int(args.table_id)
        table_id_list = [offs + x for x in range(len(args.file_list))]
    else:
        table_id_list = [int(x) for x in table_id_list]
    if len(table_id_list) != len(args.file_list):
        raise KeyError("you should provide one table_id per binning table")

    # read lineskip data
    lineskip_data = np.loadtxt(args.lineskip, delimiter=",", dtype=np.uint8)
    if lineskip_data.ndim == 1:
        lineskip_data = lineskip_data[np.newaxis, :]
    elif lineskip_data.shape[0] > len(args.file_list):
        lineskip_data = lineskip_data[1:, :]

    # create BinningTable object
    bin_ckd = BinningTables()
    bin_ckd.create_if_needed(args.validity_start)

    # add binning tables
    for ii, flname in enumerate(args.file_list):
        flpath = Path(flname)
        if not flpath.is_file():
            raise FileNotFoundError(flname)
        if flpath.suffix != ".csv":
            continue

        # read csv-file
        table = np.loadtxt(flpath, delimiter=",", dtype=np.uint32)

        # write new binning-table
        bin_ckd.add_table(table_id_list[ii], lineskip_data[ii, :], table)


# --------------------------------------------------
if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Python implementation SPEXone instrument simulator output to L1A.

Environment::

   CKD_DIR:  directory with SPEXone CKD, default is CWD.
"""

from __future__ import annotations

import argparse
from datetime import UTC, datetime, time
from os import environ
from pathlib import Path

import h5py
import numpy as np
import xarray as xr
from pyspex.l1a_io import L1Aio
from pyspex.lib import pyspex_version
from pyspex.lib.tmtc_def import tmtc_dtype
from pyspex.lv1_gse import LV1gse
from pyspex.tif_io import TIFio

# - global parameters ------------------------------
EPOCH = datetime(1970, 1, 1, tzinfo=UTC)


# - local functions --------------------------------
def get_table_id(ckd_dir: str, bin_table_name: str) -> str:
    """Quick implementation to obtain table_id from binning-table CKD."""
    if not Path(ckd_dir).is_dir():
        ckd_dir = environ.get("CKD_DIR", ".")

    # only read the latest version of the binning-table CKD
    bin_tables = sorted(list(Path(ckd_dir).glob("SPX1_OCAL_L1A_TBL_*.nc")))
    if not bin_tables:
        raise FileNotFoundError("No binning table found")

    table_id = None
    with h5py.File(bin_tables[-1], "r") as fid:
        for grp in [x for x in fid if x.startswith("Table")]:
            gid = fid[grp]
            if "origin" not in gid.attrs:
                continue

            if gid.attrs["origin"].decode("ascii") == bin_table_name:
                table_id = int(grp.split("_")[1])
                break

    return table_id


def get_stimulus(hdr: np.ndarray) -> xr.Dataset:
    """Return stimulus data as a xarray::Dataset."""
    if "Spectral data stimulus" not in hdr:
        return None

    ds_dict = hdr["Spectral data stimulus"]
    xr_wv = xr.DataArray(
        ds_dict["Wavelength (nm)"],
        coords=[ds_dict["Wavelength (nm)"]],
        dims=["wavelength"],
        attrs={"long_name": "wavelength of stimulus", "units": "nm"},
    )

    data = ds_dict["Radiance (photons/(s.nm.m^2.sr)"]
    data *= 2.99792458 * 6.62607015e-14 / ds_dict["Wavelength (nm)"]
    xr_sign = xr.DataArray(
        data,
        dims=["wavelength"],
        coords=[ds_dict["Wavelength (nm)"]],
        attrs={"long_name": "signal of stimulus", "units": "W.m-2.sr-1.um-1"},
    )

    return xr.Dataset({"wavelength": xr_wv, "signal": xr_sign})


def get_l1a_name(msm_id: str, utc_sensing_start: datetime) -> str:
    """Return name of SPEXone product for DEM measurements.

    Parameters
    ----------
    msm_id : string, optional
       Provide identifier for measurement, OCAL only
    utc_sensing_start: datetime
       Provide sensing start of first measurement in L1A product

    Notes
    -----
    L1A file name format:
       SPX1_OCAL_<msm_id>[_YYYYMMDDTHHMMSS]_L1A_vvvvvvv.nc
    where
       msm_id is the measurement identifier
       YYYYMMDDTHHMMSS is time stamp of the first image in the file
       vvvvvvv is the git-hash string of the pyspex repository

    """
    # define string of sensing start as yyyymmddThhmmss
    sensing_start = utc_sensing_start.strftime("%Y%m%dT%H%M%S")

    return (
        f"SPX1_OCAL_{msm_id}_{sensing_start}" f"_L1A_{pyspex_version(githash=True)}.nc"
    )


# - main function ----------------------------------
def main() -> None:
    """Create a L1A calibration product from data in Tiff format.

    Environment
    -----------
    CKD_DIR :  directory with SPEXone CKD, default is CWD
    """
    parser = argparse.ArgumentParser(
        description="create SPEXone L1A product from instrument simulations"
    )
    parser.add_argument(
        "--lineskip",
        default=False,
        action="store_true",
        help="read line-skip data, when available",
    )
    parser.add_argument(
        "--output", default=None, help="define output directory, default=CWD"
    )
    inp_grp = parser.add_argument_group("inp_grp", "process inp_tif files")
    inp_grp.add_argument(
        "--inp_tif",
        default=False,
        action="store_true",
        help="use TIFF file with 32-bit image",
    )
    inp_grp.add_argument(
        "--nframe",
        default=None,
        type=int,
        help="specify the number of images to be generated",
    )
    parser.add_argument(
        "--verbose", default=False, action="store_true", help="be verbose"
    )
    parser.add_argument(
        "ascii_file",
        default=None,
        help=(
            "provide path to SPEXone instrument simulator"
            ' product with extension ".dat"'
        ),
    )
    args = parser.parse_args()
    if args.verbose:
        print(args)

    if not Path(args.ascii_file).is_file():
        raise FileNotFoundError(f"File {args.ascii_file} does not exist")

    if not args.inp_tif and args.nframe is not None:
        raise RuntimeError('option --nframe works only with "--inp_tif"')

    tif = TIFio(args.ascii_file, inp_tif=args.inp_tif, lineskip=args.lineskip)
    hdr = tif.header()
    tags = tif.tags()[0]
    images = tif.images(n_frame=args.nframe)
    if args.verbose:
        for key, value in hdr.items():
            if key == "Spectral data stimulus":
                continue
            print(f"{key:s}: '{value}'")

    # For now, we obtain the reference time from the TIFF.tags
    utc_start = None
    msm_id = hdr["OCAL measurement"]
    for key in tags:
        if key.name == "datetime":
            str_date, str_time = tags[key.value].split(" ")
            utc_start = datetime.fromisoformat(
                str_date.replace(":", "-") + "T" + str_time + "+00:00"
            )
            break

    prod_name = get_l1a_name(msm_id, utc_start)
    if args.inp_tif:
        prod_name = prod_name.replace("_L1A_", "_inp_L1A_")
    if args.output is not None:
        dest_dir = Path(args.output)
        if not dest_dir.is_dir():
            dest_dir.mkdir()
        prod_name = str(dest_dir / prod_name)
    if args.verbose:
        print(prod_name)

    # define dimensions for L1A product
    if args.inp_tif:
        if images.ndim == 2:
            n_images = 1
            n_samples = images.size
        else:
            n_images = images.shape[0]
            n_samples = images.shape[1] * images.shape[2]
    else:
        n_images = int(hdr["Number of measurements"])
        if args.lineskip:
            n_samples = int(hdr["Enabled lines"]) * int(hdr["Column dimension"])
        else:
            n_samples = int(hdr["Row dimension"]) * int(hdr["Column dimension"])

    dims = {
        "number_of_images": n_images,
        "samples_per_image": n_samples,
        "hk_packets": n_images,
        "viewing_angles": int(hdr["Number of viewing angles"]),
    }

    # define binning table ID
    if n_samples == 0x400000:
        table_id = 0
    elif n_samples == 0x100000:
        table_id = 1
    else:
        if args.lineskip:
            table_id = get_table_id("/nfs/SPEXone/share/ckd", hdr["Line-enable-array"])
        else:
            table_id = get_table_id(
                "/nfs/SPEXone/share/ckd", hdr["Flexible binning table"]
            )

    # Compute delta_time for each frame (seconds)
    # convert timestamps to seconds per day
    offs = (utc_start - EPOCH).total_seconds()
    intg_time = int(hdr["Co-additions"]) * float(hdr["Exposure time (s)"])
    secnds = (1 + np.arange(n_images, dtype=float)) * intg_time
    img_sec = (offs + secnds // 1).astype("u4")
    img_subsec = np.round((secnds % 1) * 2**16).astype("u2")
    img_time = secnds

    nom_hk = np.zeros(n_images, dtype=tmtc_dtype(0x320))
    sci_hk = np.zeros(n_images, dtype=tmtc_dtype(0x350))
    if table_id == 0:
        sci_hk["REG_FULL_FRAME"] = 1
        sci_hk["REG_CMV_OUTPUTMODE"] = 3
    else:
        sci_hk["REG_FULL_FRAME"] = 2
        sci_hk["REG_CMV_OUTPUTMODE"] = 1
    sci_hk["REG_BINNING_TABLE_START"] = 0x400000 * (table_id + 1) + 0x80000000
    sci_hk["DET_OFFSET"] = 16384 - 70
    sci_hk["DET_FOTLEN"] = 20
    sci_hk["DET_EXPTIME"] = int(1e7 * float(hdr["Exposure time (s)"]) / 129 - 0.43 * 20)
    sci_hk["REG_NCOADDFRAMES"] = int(hdr["Co-additions"])

    # Generate L1A product
    ref_date = datetime.combine(utc_start.date(), time(0), UTC)
    with L1Aio(prod_name, dims=dims, ref_date=ref_date) as l1a:
        # write image data, detector telemetry and image attributes
        l1a.fill_science(images.reshape(n_images, -1), sci_hk, np.arange(n_images))

        l1a.set_dset("/image_attributes/icu_time_sec", img_sec)
        l1a.set_attr(
            "valid_min", np.uint32(1577800000), ds_name="/image_attributes/icu_time_sec"
        )
        l1a.set_attr(
            "valid_max", np.uint32(1735700000), ds_name="/image_attributes/icu_time_sec"
        )
        l1a.set_attr(
            "units",
            "seconds since 1970-01-01 00:00:00",
            ds_name="/image_attributes/icu_time_sec",
        )
        l1a.set_dset("/image_attributes/icu_time_subsec", img_subsec)
        l1a.set_dset("/image_attributes/image_time", img_time)

        # Engineering data
        l1a.fill_nomhk(nom_hk)
        l1a.set_dset("/engineering_data/HK_tlm_time", img_time)

        # Global attributes
        l1a.fill_global_attrs(inflight=False)
        l1a.set_attr("input_files", [Path(args.ascii_file).name])

    with LV1gse(prod_name) as gse:
        gse.set_attr("origin", hdr["history"])
        gse.set_attr("comment", hdr["OCAL measurement"])
        if "Light source" in hdr:
            gse.set_attr("light_source", hdr["Light source"])
        if "Detector illumination" in hdr:
            gse.set_attr("Light_source", hdr["Detector illumination"])
            gse.set_attr("Illumination_level", float(hdr["Detect. illumination e/ms"]))
        if "Detector temperature (K)" in hdr:
            gse.set_attr("Detector_temperature", float(hdr["Detector temperature (K)"]))
        if "Detector temperature (K)" in hdr:
            gse.set_attr("Optics_temperature", float(hdr["Optics temperature (K)"]))
        if "Illuminated viewing angle" in hdr:
            vp_dict = {"0": 1, "1": 2, "2": 16, "3": 8, "4": 4}
            gse.write_viewport(vp_dict.get(hdr["Illuminated viewing angle"]))
        else:
            gse.write_viewport(0)
        if "Illuminated field ACT" in hdr:
            gse.write_attr_act(
                float(hdr["Rotation stage ACT angle"]),
                float(hdr["Illuminated field ACT"]),
            )
        else:
            gse.write_attr_act(float(hdr["Rotation stage ACT angle"]))
        if "Illuminated field ALT" in hdr:
            gse.write_attr_alt(
                float(hdr["Rotation stage ALT angle"]),
                float(hdr["Illuminated field ALT"]),
            )
        else:
            gse.write_attr_alt(float(hdr["Rotation stage ALT angle"]))
        if "DoLP input" in hdr:
            gse.write_attr_polarization(
                float(hdr["AoLP input (deg)"]), float(hdr["DoLP input"])
            )
        if "Spectral data stimulus" in hdr and "Illuminated viewing angle" in hdr:
            gse.write_data_stimulus(get_stimulus(hdr))


# --------------------------------------------------
if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2022-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Script to copy an SPEXone level-1A product to a new level-1A product."""

from __future__ import annotations

import argparse
from datetime import UTC, datetime
from pathlib import Path

import h5py
import numpy as np
from pyspex.l1a_io import L1Aio

DEFAULT_EPOCH = datetime(1970, 1, 1, tzinfo=UTC)


# --------------------------------------------------
def inv_sec_of_day(
    reference_day: datetime, sec_of_day: np.ndarray, epoch: datetime = DEFAULT_EPOCH
) -> tuple[np.ndarray, np.ndarray]:
    """Convert seconds after midnight to CCSDS timestamps.

    Parameters
    ----------
    reference_day :  datetime object
       `date` object represent a date (year, month, day)
    sec_of_day :  ndarray
       Seconds since midnight on `reference_day`
    epoch :  datetime object, optional
       Provide timestamp of epoch, default 1970-01-01

    Returns
    -------
    tuple holding CCSDS timestamps

    """
    offs = np.uint32((reference_day - epoch).total_seconds())
    ccsds_sec = sec_of_day.astype("u4")
    ccsds_subsec = (65536 * (sec_of_day - ccsds_sec)).astype("u2")

    return offs + ccsds_sec, ccsds_subsec


# --------------------------------------------------
def main() -> None:
    """Copy data from one SPEXone L1A product to a new L1A product."""
    parser = argparse.ArgumentParser(
        description=(
            "Copy selected data from one SPEXone L1A product"
            " into a new SPEXone L1A product"
        )
    )
    parser.add_argument(
        "--verbose", "-v", action="store_true", help="be verbose, default be silent"
    )
    parser.add_argument("--mps_id", type=int, default=None, help="select on MPS-ID")
    # parser.add_argument('--mon_type', default=None,
    #                    help=('Specify monitoring type identifier: '
    #                          'MON-DARK, MON-NOISE, MON-NLIN, ...'))
    parser.add_argument(
        "--out",
        default=".",
        help=(
            "name of directory to store the new Level-1A"
            " product, default: current working directory"
        ),
    )
    parser.add_argument(
        "l1a_product", default=None, help="name of SPEXone Level-1A product"
    )
    args = parser.parse_args()
    if args.verbose:
        print(args)

    l1a_product = Path(args.l1a_product)
    if not l1a_product.is_file():
        raise FileNotFoundError(f"File {args.l1a_product} does not exist")
    # ToDo: check if SPEXone Level-1a product

    out_dir = Path(args.out)
    if not out_dir.is_dir():
        out_dir.mkdir(mode=0o755, parents=True)

    # ----- read data from original product -----
    # pylint: disable=no-member, unsubscriptable-object
    with h5py.File(l1a_product) as fid:
        # read image data, detector telemetry and image attributes
        # - datasets: img_data, img_hk, img_id, img_sec, img_subsec
        img_data = fid["/science_data/detector_images"][:]
        img_hk = fid["/science_data/detector_telemetry"][:]
        img_id = fid["/image_attributes/image_ID"][:]
        img_sec = fid["/image_attributes/icu_time_sec"][:]
        img_subsec = fid["/image_attributes/icu_time_subsec"][:]
        img_time = fid["/image_attributes/image_time"][:]
        # obtain reference date
        units = fid["/image_attributes/image_time"].attrs["units"]
        reference_day = datetime.fromisoformat(units[14:].decode("ascii"))

        # read engineering data
        # - datasets: nomhk_data, nomhk_sec, nomhk_subsec
        nomhk_time = None
        nomhk_data = fid["/engineering_data/NomHK_telemetry"][:]
        if nomhk_data.size > 0:
            nomhk_time = fid["/engineering_data/HK_tlm_time"][:]

        # read navigation data

        # read additional attributes:
        inflight = not fid.attrs["institution"].startswith(b"SRON")
        input_files = fid.attrs["input_files"]

    # ----- perform data selection -----
    name = l1a_product.name
    if args.mps_id is not None:
        mask = img_hk["MPS_ID"] == args.mps_id
        img_data = img_data[mask, :]
        img_hk = img_hk[mask]
        img_id = img_id[mask]
        img_sec = img_sec[mask]
        img_subsec = img_subsec[mask]
        img_time = img_time[mask]

        if nomhk_data.size > 0:
            mask = nomhk_data["MPS_ID"] == args.mps_id
            nomhk_time = nomhk_time[mask]
            nomhk_data = nomhk_data[mask]

        if name.find("PACE_SPEXone_CAL") != -1:
            name = name.replace(
                "PACE_SPEXone_CAL", f"PACE_SPEXone_CAL_mps{args.mps_id:03d}"
            )
        else:
            name = name.replace("PACE_SPEXone", f"PACE_SPEXone_mps{args.mps_id:03d}")
    l1a_filename = out_dir / name

    # define dimensions
    dims = {
        "number_of_images": img_data.shape[0],
        "samples_per_image": img_data.shape[1],
        "hk_packets": nomhk_data.size,
    }

    # ----- now we can update the name of the output product -----
    # - because the production time has changed
    # - and when coverage time is changed
    if l1a_filename.is_file() and l1a_product.samefile(l1a_filename):
        raise OSError("Output will overwrite original product")

    # ----- write new output product with selected data -----
    with L1Aio(str(l1a_filename), reference_day, dims=dims) as l1a:
        # write image data, detector telemetry and image attributes
        l1a.fill_science(img_data, img_hk, img_id)
        l1a.set_dset("/image_attributes/icu_time_sec", img_sec)
        l1a.set_dset("/image_attributes/icu_time_subsec", img_subsec)
        l1a.set_dset("/image_attributes/image_time", img_time)

        # write engineering data
        if nomhk_data.size > 0:
            l1a.fill_nomhk(nomhk_data)
            l1a.set_dset("/engineering_data/HK_tlm_time", nomhk_time)

        # write navigation data

        # write global attributes
        l1a.fill_global_attrs(inflight=inflight)
        l1a.set_attr("input_files", input_files.tolist())

    # copy group with EGSE/OGSE data
    # *** DO NOT USE: BREAKS NETCDF4 FORMAT ***
    # if gse_data:
    #     print('copy EGSE/OGSE data')
    #     with h5py.File(l1a_product) as fid_in:
    #         with h5py.File(l1a_filename, 'r+') as fid_out:
    #             fid_out.copy(fid_in['gse_data'], 'gse_data')


# --------------------------------------------------
if __name__ == "__main__":
    main()



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Contains the class `HKTio` to read PACE HKT products."""

from __future__ import annotations

__all__ = ["CoverageFlag", "HKTio"]

import logging
from enum import IntFlag, auto
from pathlib import Path
from typing import TYPE_CHECKING

import h5py
import numpy as np
import xarray as xr

from .lib.ccsds_hdr import CCSDShdr

if TYPE_CHECKING:
    import datetime as dt

# - global parameters -----------------------
module_logger = logging.getLogger("pyspex.hkt_io")

# reduce extend of navigation data to
TEN_SECONDS = np.timedelta64(10, "s")
# require an extent of navigation data of
SEVEN_SECONDS = np.timedelta64(7, "s")


# - class CoverageFlag ----------------------
class CoverageFlag(IntFlag):
    """Define flags for coverage_quality (navigation_data)."""

    MISSING_SAMPLES = auto()
    TOO_SHORT_EXTENDS = auto()
    NO_EXTEND_AT_START = auto()
    NO_EXTEND_AT_END = auto()

    @classmethod
    def check(
        cls: CoverageFlag,
        nav_data: xr.Dataset,
        coverage: list[dt.datetime, dt.datetime],
    ) -> int:
        """Check coverage time of navigation data."""
        coverage_quality = cls(0)
        l1a_time_range = (
            np.datetime64(coverage[0].replace(tzinfo=None)),
            np.datetime64(coverage[1].replace(tzinfo=None)),
        )
        nav_time_range = (
            nav_data["att_time"].values[0],
            nav_data["att_time"].values[-1],
        )

        # check at the start of the data
        if l1a_time_range[0] < nav_time_range[0]:
            coverage_quality |= CoverageFlag.NO_EXTEND_AT_START
            module_logger.error(
                "time coverage of navigation data starts after L1A science data"
            )

        diff_coverage = l1a_time_range[0] - nav_time_range[0]
        if diff_coverage <= SEVEN_SECONDS:
            coverage_quality |= CoverageFlag.TOO_SHORT_EXTENDS
            module_logger.warning(
                "extends of time coverage of navigation data too short",
            )

        # check at the end of the data
        if l1a_time_range[1] > nav_time_range[1]:
            coverage_quality |= CoverageFlag.NO_EXTEND_AT_END
            module_logger.error(
                "time coverage of navigation data ends before L1A science data"
            )

        diff_coverage = nav_time_range[1] - l1a_time_range[1]
        if diff_coverage <= SEVEN_SECONDS:
            coverage_quality |= CoverageFlag.TOO_SHORT_EXTENDS
            module_logger.warning(
                "extends of time coverage of navigation data too short",
            )
        return coverage_quality


# - class HKTio -------------------------
class HKTio:
    """Class to read housekeeping and navigation data from PACE-HKT products.

    Parameters
    ----------
    flnames :  str | Path | list[Path]
        sorted list of PACE_HKT filenames (netCDF4 format)

    Notes
    -----
    This class has the following methods::

     - housekeeping(instrument: str) -> tuple[np.ndarray]
     - navigation() -> dict

    """

    def __init__(self: HKTio, flnames: str | Path | list[Path]) -> None:
        """Initialize access to PACE HKT products."""
        self.nav_data: xr.Dataset | None = None
        if isinstance(flnames, str | Path):
            self.flnames = [flnames] if isinstance(flnames, Path) else [Path(flnames)]
        else:
            self.flnames = flnames

        for hkt_fl in self.flnames:
            if not hkt_fl.is_file():
                raise FileNotFoundError(f"file {hkt_fl} not found on system")

    @staticmethod
    def __read_hkt__(hkt_file: Path, instrument: str) -> np.ndarray | None:
        """Return housekeeping data of a given instrument."""
        with h5py.File(hkt_file) as fid:
            gid = fid["housekeeping_data"]
            if instrument == "spx":
                ds_name = (
                    "SPEX_HKT_packets"
                    if "SPEX_HKT_packets" in gid
                    else "SPEXone_HKT_packets"
                )
            elif instrument == "sc":
                ds_name = "SC_HKT_packets"
            elif instrument == "oci":
                ds_name = "OCI_HKT_packets"
            elif instrument == "harp":
                ds_name = (
                    "HARP_HKT_packets"
                    if "HARP_HKT_packets" in gid
                    else "HARP2_HKT_packets"
                )
            else:
                raise KeyError("data of unknown instrument requested")

            res = gid[ds_name][:] if ds_name in gid else None

        return res

    # ---------- PUBLIC FUNCTIONS ----------
    def housekeeping(self: HKTio, instrument: str = "spx") -> tuple[np.ndarray]:
        """Get housekeeping telemetry data.

        Parameters
        ----------
        instrument : {'spx', 'oci', 'harp', 'sc'}, default='spx'
           name of PACE instrument: 'harp': HARP2, 'oci': OCI,
           'sc': spacecraft, 'spx': SPEXone.

        Notes
        -----
        Current implementation only works for SPEXone.

        """
        res = None
        for hkt_file in self.flnames:
            buff = self.__read_hkt__(hkt_file, instrument)
            if buff is not None:
                res = buff if res is None else np.concatenate((res, buff), axis=0)
        if res is None:
            return ()

        ccsds_hk = ()
        for packet in res:
            try:
                ccsds_hdr = CCSDShdr()
                ccsds_hdr.read("raw", packet)
            except ValueError as exc:
                module_logger.warning('CCSDS header read error with "%s"', exc)
                break

            try:
                dtype_apid = ccsds_hdr.data_dtype
            except ValueError:
                print(
                    f"APID: 0x{ccsds_hdr.apid:x};"
                    f" Packet Length: {ccsds_hdr.packet_size:d}"
                )
                dtype_apid = None

            if dtype_apid is not None:  # all valid APIDs
                buff = np.frombuffer(packet, count=1, offset=0, dtype=dtype_apid)
                ccsds_hk += (buff,)
            else:
                module_logger.warning(
                    "package with APID 0x%x and length %d is not implemented",
                    ccsds_hdr.apid,
                    ccsds_hdr.packet_size,
                )

        return ccsds_hk

    def navigation(self: HKTio) -> None:
        """..."""
        if self.nav_data is not None:
            self.nav_data.close()
            self.nav_data = None

        nav = None
        for hkt_file in self.flnames:
            if nav is None:
                nav = xr.open_dataset(hkt_file, group="navigation_data")
            else:
                res = ()
                buff = xr.open_dataset(hkt_file, group="navigation_data")
                for key, xarr in buff.data_vars.items():
                    res += (xr.concat((nav[key], buff[key]), dim=xarr.dims[0]),)
                nav = xr.merge(res, combine_attrs="drop_conflicts")

        # fix coordinates
        if "att_time" in nav.data_vars:
            nav = nav.rename({"att_records": "att_time"}).set_coords(["att_time"])
        if "orb_time" in nav.data_vars:
            nav = nav.rename({"orb_records": "orb_time"}).set_coords(["orb_time"])
        if "tilt_time" in nav.data_vars:
            nav = nav.rename({"tilt_records": "tilt_time"}).set_coords(["tilt_time"])
        # clean-up Dataset attributes
        for key in list(nav.attrs):
            del nav.attrs[key]
        nav.attrs["time_coverage_start"] = str(nav["att_time"].values[0])[:23] + "Z"
        nav.attrs["time_coverage_end"] = str(nav["att_time"].values[-1])[:23] + "Z"
        self.nav_data = nav

    def add_nav(
        self: HKTio, l1a_file: Path, coverage: list[dt.datetime, dt.datetime]
    ) -> None:
        """..."""
        time_range = (
            np.datetime64(coverage[0].replace(tzinfo=None)) - TEN_SECONDS,
            np.datetime64(coverage[1].replace(tzinfo=None)) + TEN_SECONDS,
        )
        att_indx = (
            (self.nav_data.att_time.values >= time_range[0])
            & (self.nav_data.att_time.values <= time_range[1])
        ).nonzero()[0]
        orb_indx = (
            (self.nav_data.orb_time.values >= time_range[0])
            & (self.nav_data.orb_time.values <= time_range[1])
        ).nonzero()[0]
        tilt_indx = (
            (self.nav_data.tilt_time.values >= time_range[0])
            & (self.nav_data.tilt_time.values <= time_range[1])
        ).nonzero()[0]
        nav = self.nav_data.isel(
            att_time=att_indx, orb_time=orb_indx, tilt_time=tilt_indx
        )
        self.nav_data.close()

        # add coverage-quality flag
        xarr = xr.DataArray(
            np.array([CoverageFlag.check(nav, coverage)]),
            name="coverage_quality",
            attrs={
                "long_name": "coverage quality of navigation data",
                "standard_name": "status_flag",
                "valid_range": np.array([0, 15], dtype="u2"),
                "flag_values": np.array([0, 1, 2, 4, 8], dtype="u2"),
                "flag_meanings": (
                    "good missing-samples too_short_extends no_extend_at_start"
                    " no_extend_at_end"
                ),
            },
        )
        nav["coverage_quality"] = xarr
        nav.to_netcdf(l1a_file, group="navigation_data", mode="a")
        nav.close()



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
#
"""Tools to read or write definitions of SPEXone binning-tables."""

from __future__ import annotations

__all__ = ["BinningTables"]

from datetime import UTC, datetime
from importlib.resources import files

import numpy as np

# pylint: disable=no-name-in-module
from netCDF4 import Dataset

from .lib import pyspex_version

# - global parameters ------------------------------
FILL_VALUE = 0xFFFFFFFF  # 0X7FFFFFFF


# - local functions --------------------------------
class BinningTables:
    """Class to handle SPEXone binning-table definitions.

    Raises
    ------
    FileNotFoundError
        Directory with SPEXone binning-table files does not exist.

    Notes
    -----
    Syntax of the file name with SPEXone binning-tables:

          SPX1_CKD_BIN_TBL_<yyyymmddTHHMMSS>_<NNN>.nc

    where yyyymmddTHHMMSS defines the validity start (UTC) and NNN the
    release number of the file format.

    The binning tables as defined on-ground are supposed to be available
    during the whole mission at the same on-board memory location.
    Because these original binning tables are necessary for re-processing
    and may facilitate instrument performance monitoring.
    Therefore, it is preferred that a new binning table is added to the
    current set, without changing the validity start string.
    However, new binning-table file should be released in case any of the
    binning tables are overwritten.

    Examples
    --------
    # create new file with binning-table definitions::

    > bin_tbl = BinningTables()
    > bin_tbl.create_if_needed(validity_start)
    > bin_tbl.add_table(0, lineskip_arr, binning_table)
    > bin_tbl.add_table(1, lineskip_arr, binning_table)

    # use binning-table '130' to unbin SPEXone detector data::

    > bin_tbl = BinningTables(130)
    > img_1d = bin_tbl.unbin(img_binned)
    > img_2d = bin_tbl.to_image(img_1d)

    """

    def __init__(self: BinningTables, table_id: int | None = None) -> None:
        """Initialize class attributes."""
        self.bin_tbl = files("pyspex.data").joinpath("binning_tables.nc")
        if table_id is None:
            return

        if not self.bin_tbl.is_file():
            raise FileNotFoundError(f"{self.bin_tbl} not found")

        with Dataset(self.bin_tbl, "r") as fid:
            if f"Table_{table_id:03d}" not in fid.groups:
                raise KeyError(f"Table_{table_id:03d} not defined")
            gid = fid[f"Table_{table_id:03d}"]
            self.binning_table = gid.variables["binning_table"][:]
            self.lineskip_arr = gid.variables["lineskip_arr"][:]
            self.count_table = gid.variables["count_table"][:]

    def unbin(self: BinningTables, img_binned: np.ndarray) -> np.ndarray:
        """Return detector data corrected for flexible and fixed binning (still 1-D).

        Parameters
        ----------
        img_binned : np.ndarray
           Binned image data (1D array)

        Returns
        -------
        np.ndarray
           unbinned image data

        """
        img_binned /= 4 * self.count_table
        mask = self.count_table > 5

        if img_binned.ndim == 2:
            img_binned[:, mask] = np.nan
            return img_binned

        img_binned[mask] = np.nan
        return img_binned

    def to_image(self: BinningTables, img_1d: np.ndarray) -> np.ndarray:
        """Return unbinned detector data.

        Parameters
        ----------
        img_1d : np.ndarray
           unbinned image data (1-D array)

        Returns
        -------
        np.ndarray
           2x2 image (2-D array).

        """
        revert = np.full(self.binning_table.shape, np.nan)
        table = self.binning_table[self.lineskip_arr == 1, :].reshape(-1)
        revert[self.lineskip_arr == 1, :] = img_1d[table].reshape(-1, 1024)
        return revert

    def create_if_needed(
        self: BinningTables, validity_start: str, release: int = 1
    ) -> None:
        """Initialize CKD file for binning tables if not exist.

        Parameters
        ----------
        validity_start: str
           Validity start of the CKD data, as ``yyyymmddTHHMMSS``
        release :  int, default=1
           Release number, start at 1

        """
        # initialize netCDF file with binning tables
        with Dataset("binning_table.nc", "w") as fid:
            fid.title = "SPEXone Level-1 binning-tables"
            fid.Conventions = "CF-1.6"
            fid.project = "PACE Project"
            fid.instrument = "SPEXone"
            fid.institution = "SRON Netherlands Institute for Space Research"
            fid.processing_version = pyspex_version()
            fid.validity_start = validity_start + "+00:00"
            fid.release_number = np.uint16(release)
            fid.date_created = datetime.now(UTC).isoformat(timespec="seconds")

            fid.createDimension("row", 1024)
            fid.createDimension("column", 1024)

    def add_table(
        self: BinningTables,
        table_id: int,
        lineskip_arr: np.ndarray,
        binning_table: np.ndarray,
    ) -> None:
        """Add a binning table definition to existing file.

        Parameters
        ----------
        table_id :  int
           Table identifier (integer between 1 and 255)
        lineskip_arr :  ndarray
           Lineskip array definition
        binning_table :  ndarray
           Binning table definition

        """
        index, count = np.unique(
            binning_table[lineskip_arr == 1, :], return_counts=True
        )

        with Dataset(self.bin_tbl, "r+") as fid:
            gid = fid.createGroup(f"/Table_{table_id:03d}")
            gid.tabel_id = table_id
            gid.REG_BINNING_TABLE_START = hex(0x80000000 + 0x400000 * (table_id - 1))
            gid.enabled_lines = np.uint16(lineskip_arr.sum())
            gid.flex_binned_pixels = np.uint32(index.max() + 1)
            gid.date_created = datetime.now(UTC).isoformat(timespec="seconds")

            dset = gid.createVariable(
                "binning_table",
                "u4",
                ("row", "column"),
                fill_value=FILL_VALUE,
                chunksizes=(128, 128),
                zlib=True,
                complevel=1,
                shuffle=True,
            )
            dset.long_name = "binning table"
            dset.valid_min = np.uint32(0)
            dset.valid_max = np.uint32(index.max())
            dset[:] = binning_table

            dset = gid.createVariable(
                "lineskip_arr", "u1", ("row",), zlib=True, complevel=1, shuffle=True
            )
            dset.long_name = "lineskip array"
            dset.valid_min = np.uint8(0)
            dset.valid_max = np.uint8(1)
            dset[:] = lineskip_arr

            gid.createDimension("bins", count.size)
            dset = gid.createVariable(
                "count_table", "u2", ("bins",), zlib=True, complevel=1, shuffle=True
            )
            dset.long_name = "number of aggregated pixel readings"
            dset.valid_min = np.uint16(0)
            dset.valid_max = np.uint16(count.max())
            dset[:] = count.astype("u2")



#
# This file is part of pyspex
#
# https://github.com/rmvanhees/pyspex.git
#
# Copyright (c) 2019-2024 SRON - Netherlands Institute for Space Research
#    All Rights Reserved
#
# License:  BSD-3-Clause
"""Contains the class `L1Aio` to write PACE/SPEXone data in level-1A format."""

from __future__ import annotations

__all__ = ["L1Aio"]

import logging
from pathlib import Path, PurePosixPath
from typing import TYPE_CHECKING, Any

import numpy as np

from .lib.attrs_def import attrs_def
from .lib.l1a_def import init_l1a
from .lib.tlm_utils import convert_hk

if TYPE_CHECKING:
    import datetime as dt


# - global parameters -------------------
module_logger = logging.getLogger("pyspex.l1a_io")

MCP_TO_SEC = 1e-7
ONE_DAY = 24 * 60 * 60


# - local functions ---------------------
def _binning_table_(img_hk: np.ndarray) -> np.ndarray:
    """Return binning table identifier (zero for full-frame images)."""
    if "REG_FULL_FRAME" not in img_hk.dtype.names:
        module_logger.warning("can not determine binning table identifier")
        return np.full(len(img_hk), -1, dtype="i1")

    full_frame = np.unique(img_hk["REG_FULL_FRAME"])
    if len(full_frame) > 1:
        module_logger.warning("value of REG_FULL_FRAME not unique")
    full_frame = img_hk["REG_FULL_FRAME"][-1]

    cmv_outputmode = np.unique(img_hk["REG_CMV_OUTPUTMODE"])
    if len(cmv_outputmode) > 1:
        module_logger.warning("value of REG_CMV_OUTPUTMODE not unique")
    cmv_outputmode = img_hk["REG_CMV_OUTPUTMODE"][-1]

    if full_frame == 1:
        if cmv_outputmode != 3:
            raise KeyError("Diagnostic mode with REG_CMV_OUTPUTMODE != 3")
        return np.zeros(len(img_hk), dtype="i1")

    if full_frame == 2:
        if cmv_outputmode != 1:
            raise KeyError("Science mode with REG_CMV_OUTPUTMODE != 1")
        bin_tbl_start = img_hk["REG_BINNING_TABLE_START"]
        indx0 = np.nonzero(img_hk["REG_FULL_FRAME"] != 2)[0]
        if indx0.size > 0:
            indx2 = np.nonzero(img_hk["REG_FULL_FRAME"] == 2)[0]
            bin_tbl_start[indx0] = bin_tbl_start[indx2[0]]
        res = 1 + (bin_tbl_start - 0x80000000) // 0x400000
        return res & 0xFF

    raise KeyError("REG_FULL_FRAME not equal to 1 or 2")


def _digital_offset_(img_hk: np.ndarray) -> np.ndarray:
    """Return digital offset including ADC offset [count]."""
    buff = img_hk["DET_OFFSET"].astype("i4")
    if np.isscalar(buff):
        if buff >= 8192:
            buff -= 16384
    else:
        buff[buff >= 8192] -= 16384

    return buff + 70


def _exposure_time_(img_hk: np.ndarray) -> np.ndarray:
    """Return exposure time in seconds [float]."""
    return 129e-4 * (0.43 * img_hk["DET_FOTLEN"][:] + img_hk["DET_EXPTIME"][:])


def _nr_coadditions_(img_hk: np.ndarray) -> np.ndarray:
    """Return number of coadditions."""
    return img_hk["REG_NCOADDFRAMES"]


# - class L1Aio -------------------------
class L1Aio:
    """Class to create SPEXone level-1A products.

    Parameters
    ----------
    product :  str
       Name of the SPEXone level-1A product
    ref_date :  dt.datetime
       Date of the first detector image
    dims :  dict
       Dimensions of the datasets, default values::

          number_of_images : None     # number of image frames
          samples_per_image : 184000  # depends on binning table
          hk_packets : None           # number of HK tlm-packets

    compression : bool, default=False
       Use compression on dataset /science_data/detector_images

    """

    dset_stored = {
        "/science_data/detector_images": 0,
        "/science_data/detector_telemetry": 0,
        "/image_attributes/binning_table": 0,
        "/image_attributes/digital_offset": 0,
        "/image_attributes/nr_coadditions": 0,
        "/image_attributes/exposure_time": 0,
        "/image_attributes/icu_time_sec": 0,
        "/image_attributes/icu_time_subsec": 0,
        "/image_attributes/image_time": 0,
        "/image_attributes/image_ID": 0,
        "/engineering_data/NomHK_telemetry": 0,
        # '/engineering_data/DemHK_telemetry': 0,
        "/engineering_data/temp_detector": 0,
        "/engineering_data/temp_housing": 0,
        "/engineering_data/temp_radiator": 0,
        "/engineering_data/HK_tlm_time": 0,
    }

    def __init__(
        self: L1Aio,
        product: Path | str,
        ref_date: dt.datetime,
        dims: dict,
        compression: bool = False,
    ) -> None:
        """Initialize access to a SPEXone level-1A product."""
        self.product: Path = Path(product) if isinstance(product, str) else product

        # initialize private class-attributes
        self.__epoch = ref_date

        # initialize level-1A product
        self.fid = init_l1a(product, ref_date, dims, compression)
        for key in self.dset_stored:
            self.dset_stored[key] = 0

    def __iter__(self: L1Aio) -> None:
        """Allow iteration."""
        for attr in sorted(self.__dict__):
            if not attr.startswith("__"):
                yield attr

    def __enter__(self: L1Aio) -> L1Aio:
        """Initiate the context manager."""
        return self

    def __exit__(self: L1Aio, *args: str) -> bool:
        """Exit the context manager."""
        self.close()
        return False  # any exception is raised by the with statement.

    def close(self: L1Aio) -> None:
        """Close product and check if required datasets are filled with data."""
        if self.fid is None:
            return

        # check if at least one dataset is updated
        if self.fid.dimensions["number_of_images"].size == 0:
            self.fid.close()
            self.fid = None
            return

        # check of all required dataset their sizes
        self.check_stored(allow_empty=True)
        self.fid.close()
        self.fid = None

    # ---------- PUBLIC FUNCTIONS ----------
    @property
    def epoch(self: L1Aio) -> dt.datetime:
        """Provide epoch for SPEXone."""
        return self.__epoch

    def get_dim(self: L1Aio, name: str) -> int:
        """Get size of a netCDF4 dimension."""
        return self.fid.dimensions[name].size

    # ----- ATTRIBUTES --------------------
    def get_attr(self: L1Aio, name: str, ds_name: str | None = None) -> str | None:
        """Read data of an attribute.

        Global or attached to a group or variable.

        Parameters
        ----------
        name : str
           name of the attribute
        ds_name : str, default=None
           name of dataset to which the attribute is attached

        Returns
        -------
        scalar or array_like
           value of attribute 'name', global or attached to dataset 'ds_name'

        """
        if ds_name is None:
            res = self.fid.getncattr(name)
        else:
            if ds_name not in self.fid.groups and ds_name not in self.fid.variables:
                return None
            res = self.fid[ds_name].getncattr(name)

        if isinstance(res, bytes):
            return res.decode("ascii")

        return res

    def set_attr(
        self: L1Aio,
        name: str,
        value: Any,  # noqa: ANN401
        ds_name: str | None = None,
    ) -> None:
        """Write data to an attribute.

        Global or attached to a group or variable.

        Parameters
        ----------
        name : string
           name of the attribute
        value : scalar, array_like
           value or values to be written
        ds_name : str, default=None
           name of group or dataset to which the attribute is attached
           **Use group name without starting '/'**

        """
        if ds_name is None:
            if isinstance(value, str):
                self.fid.setncattr(name, np.bytes_(value))
            else:
                self.fid.setncattr(name, value)
        else:
            grp_name = str(PurePosixPath(ds_name).parent)
            var_name = str(PurePosixPath(ds_name).name)
            if grp_name != ".":
                if (
                    var_name not in self.fid[grp_name].groups
                    and var_name not in self.fid[grp_name].variables
                ):
                    raise KeyError(f"ds_name {ds_name} not in product")
            else:
                if (
                    var_name not in self.fid.groups
                    and var_name not in self.fid.variables
                ):
                    raise KeyError(f"ds_name {ds_name} not in product")

            if isinstance(value, str):
                self.fid[ds_name].setncattr(name, np.bytes_(value))
            else:
                self.fid[ds_name].setncattr(name, value)

    # ----- VARIABLES --------------------
    def get_dset(self: L1Aio, name: str) -> None:
        """Read data of a netCDF4 variable.

        Parameters
        ----------
        name : str
           name of dataset

        Returns
        -------
        scalar or array_like
           value of dataset 'name'

        """
        grp_name = str(PurePosixPath(name).parent)
        var_name = str(PurePosixPath(name).name)
        if grp_name != ".":
            if var_name not in self.fid[grp_name].variables:
                raise KeyError(f"dataset {name} not in level-1A product")
        else:
            if var_name not in self.fid.variables:
                raise KeyError(f"dataset {name} not in level-1A product")

        return self.fid[name][:]

    def set_dset(self: L1Aio, name: str, value: Any) -> None:  # noqa: ANN401
        """Write data to a netCDF4 variable.

        Parameters
        ----------
        name : str
           Name of level-1A dataset
        value : scalar or array_like
           Value or values to be written

        """
        value = np.asarray(value)
        grp_name = str(PurePosixPath(name).parent)
        var_name = str(PurePosixPath(name).name)
        if grp_name != ".":
            if var_name not in self.fid[grp_name].variables:
                raise KeyError(f"dataset {name} not in level-1A product")
        else:
            if var_name not in self.fid.variables:
                raise KeyError(f"dataset {name} not in level-1A product")

        self.fid[name][...] = value
        self.dset_stored[name] += 1 if value.shape == () else value.shape[0]

    # -------------------------
    def fill_global_attrs(self: L1Aio, inflight: bool = False) -> None:
        """Define global attributes in the SPEXone level-1A products.

        Parameters
        ----------
        inflight :  bool, default=False
           Measurements performed on-ground or inflight

        """
        dict_attrs = attrs_def(inflight)
        dict_attrs["product_name"] = self.product.name
        for key, value in dict_attrs.items():
            if value is not None:
                self.fid.setncattr(key, value)

    # - L1A specific functions ------------------------
    def check_stored(self: L1Aio, allow_empty: bool = False) -> None:
        """Check variables with the same first dimension have equal sizes.

        Parameters
        ----------
        allow_empty :  bool, default=False
           Allow variables to be empty

        """
        warn_str = (
            "SPEX level-1A format check [WARNING]:"
            ' size of variable "{:s}" is wrong, only {:d} elements'
        )

        # check image datasets
        dim_sz = self.get_dim("number_of_images")
        res = []
        key_list = [
            x
            for x in self.dset_stored
            if (x.startswith("/science_data") or x.startswith("/image_attributes"))
        ]
        for key in key_list:
            res.append(self.dset_stored[key])
        res = np.array(res)
        if allow_empty:
            indx = ((res > 0) & (res != dim_sz)).nonzero()[0]
        else:
            indx = np.nonzero(res != dim_sz)[0]
        for ii in indx:
            print(warn_str.format(key_list[ii], res[ii]))

        # check house-keeping datasets
        dim_sz = self.get_dim("hk_packets")
        key_list = [x for x in self.dset_stored if x.startswith("/engineering_data")]
        res = []
        for key in key_list:
            res.append(self.dset_stored[key])
        res = np.array(res)
        if allow_empty:
            indx = np.nonzero((res > 0) & (res != dim_sz))[0]
        else:
            indx = np.nonzero(res != dim_sz)[0]
        for ii in indx:
            print(warn_str.format(key_list[ii], res[ii]))

    # ---------- PUBLIC FUNCTIONS ----------
    def fill_science(
        self: L1Aio, img_data: np.ndarray, img_hk: np.ndarray, img_id: np.ndarray
    ) -> None:
        """Write Science data and housekeeping telemetry (Science).

        Parameters
        ----------
        img_data : numpy array (uint16)
           Detector image data
        img_hk : numpy array ()
           Structured array with all Science telemetry parameters
        img_id : numpy array (uint16)
           Detector frame counter modulo 0x3FFF

        Notes
        -----
        Adds detector_telemetry data to the group /science_data

        Parameters: binning_table, digital_offset, exposure_time
        and nr_coadditions are extracted from the telemetry packets and writen
        in the group /image_attributes

        """
        if len(img_hk) == 0:
            return

        self.set_dset("/science_data/detector_images", img_data)
        self.set_dset("/science_data/detector_telemetry", img_hk)
        self.set_dset("/image_attributes/image_ID", img_id)
        self.set_dset("/image_attributes/binning_table", _binning_table_(img_hk))
        self.set_dset("/image_attributes/digital_offset", _digital_offset_(img_hk))
        self.set_dset("/image_attributes/exposure_time", _exposure_time_(img_hk))
        self.set_dset("/image_attributes/nr_coadditions", _nr_coadditions_(img_hk))

    def fill_nomhk(self: L1Aio, nomhk_data: np.ndarray) -> None:
        """Write nominal house-keeping telemetry packets (NomHK).

        Parameters
        ----------
        nomhk_data : numpy array
           Structured array with all NomHK telemetry parameters

        Notes
        -----
        Writes nomhk_data as TM_telemetry in group /engineering_data

        Parameters: temp_detector and temp_housing are extracted and converted
        to Kelvin and writen to the group /engineering_data

        """
        if len(nomhk_data) == 0:
            return

        self.set_dset("/engineering_data/NomHK_telemetry", nomhk_data)

        if np.all(nomhk_data["TS1_DEM_N_T"] == 0):
            self.set_dset(
                "/engineering_data/temp_detector", np.full(nomhk_data.size, 273)
            )
        else:
            self.set_dset(
                "/engineering_data/temp_detector",
                convert_hk("TS1_DEM_N_T", nomhk_data["TS1_DEM_N_T"]),
            )

        if np.all(nomhk_data["TS2_HOUSING_N_T"] == 0):
            self.set_dset(
                "/engineering_data/temp_housing", np.full(nomhk_data.size, 293)
            )
        else:
            self.set_dset(
                "/engineering_data/temp_housing",
                convert_hk("TS2_HOUSING_N_T", nomhk_data["TS2_HOUSING_N_T"]),
            )

        if np.all(nomhk_data["TS3_RADIATOR_N_T"] == 0):
            self.set_dset(
                "/engineering_data/temp_radiator", np.full(nomhk_data.size, 294)
            )
        else:
            self.set_dset(
                "/engineering_data/temp_radiator",
                convert_hk("TS3_RADIATOR_N_T", nomhk_data["TS3_RADIATOR_N_T"]),
            )
